import copy
import numpy as np
from scipy import interpolate as sp
from scipy.spatial.transform import Rotation as R

import open3d as o3d
from . import base_types as base_types
from .electromagnetics import beamforming as BM
from .electromagnetics import empropagation as EM
from .geometry import geometryfunctions as GF
from .raycasting import rayfunctions as RF
from .utility import math_functions as MF
#from .models.frequency_domain import calculate_farfield as farfield_generator
#from .models.frequency_domain import calculate_scattering as frequency_domain_channel

class object3d:
    def __init__(self):
        #define object pose as rotation matrix from world frame using Denavit-Hartenbeg convention
        self.pose=np.eye(4)
    def rotate_euler(self,x,y,z,degrees=True,replace=True):
        rot=R.from_euler('xyz',[x,y,z],degrees=degrees)
        transform=np.eye(4)
        transform[:3,:3]=rot.as_matrix()
        if replace==True:
            self.pose=transform
        else:
            self.pose=np.matmul(self.pose,transform)
        return self.pose

    def rotate_matrix(self,new_axes,replace=True):
        rot=R.from_matrix(new_axes)
        transform=np.eye(4)
        transform[:3,:3]=rot.as_matrix()
        if replace==True:
            self.pose=transform
        else:
            self.pose=np.matmul(self.pose,transform)
        return self.pose


class points(object3d):
    """
    Structure class to store information about the geometry and materials in the environment, holding the seperate
    shapes as :class:`open3d.geometry.TriangleMesh` data structures. Everything in the class will be considered an integrated unit, rotating and moving together.
    This class will be developed to include material parameters to enable more complex modelling.

    Units should be SI, metres

    This is the default class for passing structures to the different models.
    """

    def __init__(self, points=None):
        super().__init__()
        # solids is a list of open3D :class:`open3d.geometry.TriangleMesh` structures
        if points==None:
            #no points provided at creation,
            print("Empty Object Created, please add points")
            self.points=[]
        else:
            self.points = []
            for item in points:
                self.points.append(item)
            # self.materials = []
            # for item in material_characteristics:
            #    self.materials.append(item)

    def remove_points(self, deletion_index):
        """
        removes a component or components from the class

        Parameters
        -----------
        deletion_index : list
            list of integers or numpy array of integers to the solids to be removed

        Returns
        --------
        None
        """
        for entry in range(len(deletion_index)):
            self.points.pop(deletion_index[entry])
            self.materials.pop(deletion_index[entry])

    def add_points(self, new_points):
        """
        adds a component or components from the structure

        Parameters
        -----------
        new_points : :class:`open3d.geometry.PointCloud`
            the point cloud to be added to the point cloud collection

        Returns
        --------
        None

        """
        self.points.append(new_points)
        # self.materials.append(new_materials)
    def create_points(self, points,normals):
        """
        create points within the class based upon the provided numpy arrays of floats in local coordinates

        Parameters
        ----------
        points : numpy 2d array
            the coordinates of all the poitns
        normals : numpy 2d array
            the normal vectors of each point

        Returns
        -------
        None
        """
        new_point_cloud=o3d.geometry.PointCloud()
        new_point_cloud.points=o3d.utility.Vector3dVector(points.reshape(-1,3))
        new_point_cloud.normals = o3d.utility.Vector3dVector(normals.reshape(-1, 3))
        self.add_points(new_point_cloud)

    def rotate_points(
        self, rotation_matrix, rotation_centre=np.zeros((3, 1), dtype=np.float32)
    ):
        """
        rotates the components of the structure around a common point, default is the origin

        Parameters
        ----------
        rotation_matrix : open3d rotation matrix
            o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector)
        rotation_centre : 1*3 numpy float array
            centre of rotation for the structures

        Returns
        --------
        None
        """
        # warning, current commond just rotates around the origin, and until Open3D can be brought up to the
        # latest version without breaking BlueCrystal reqruiements, this will require additional code.
        for item in range(len(self.points)):
            self.point[item] = GF.open3drotate(
                self.point[item], rotation_matrix, rotation_centre
            )

    def translate_points(self, vector):
        """
        translates the point clouds in the class by the given cartesian vector (x,y,z)

        Parameters
        -----------
        vector : 1*3 numpy array of floats
            The desired translation vector for the structures

        Returns
        --------
        None
        """
        for item in range(len(self.points)):
            self.points[item].translate(vector)

    def export_points(self,point_index=None):
        """
        combines all the points in the collection as a combined point cloud for modelling

        Returns
        -------
        combined points
        """
        if point_index==None:
            combined_points = o3d.geometry.PointCloud()
            for item in range(len(self.points)):
                combined_points = combined_points + self.points[item]

            combined_points.transform(self.pose)

        else:
            combined_points = o3d.geometry.PointCloud()
            for item in point_index:
                combined_points = combined_points + self.points[item]

            combined_points.transform(self.pose)

        return combined_points


class structures(object3d):
    """
    Structure class to store information about the geometry and materials in the environment, holding the seperate
    shapes as :class:`open3d.geometry.TriangleMesh` data structures. Everything in the class will be considered an integrated unit, rotating and moving together.
    This class will be developed to include material parameters to enable more complex modelling.

    Units should be SI, metres

    This is the default class for passing structures to the different models.
    """

    def __init__(self, solids=None):
        super().__init__()
        # solids is a list of open3D :class:`open3d.geometry.TriangleMesh` structures
        if solids == None:
            # no points provided at creation,
            print("Empty Object Created, please add solids")
            self.solids = []
        else:
            self.solids = []
            for item in range(len(solids)):
                self.solids.append(solids[item])
            # self.materials = []
            # for item in material_characteristics:
            #    self.materials.append(item)

    def remove_structure(self, deletion_index):
        """
        removes a component or components from the class

        Parameters
        -----------
        deletion_index : list
            list of integers or numpy array of integers to the solids to be removed

        Returns
        --------
        None
        """
        for entry in range(len(deletion_index)):
            self.solids.pop(deletion_index[entry])
            self.materials.pop(deletion_index[entry])

    def add_structure(self, new_solids):
        """
        adds a component or components from the structure

        Parameters
        -----------
        new_solids : :class:`open3d.geometry.TriangleMesh`
            the solid to be added to the structure

        Returns
        --------
        None

        """
        self.solids.append(new_solids)
        # self.materials.append(new_materials)

    def rotate_structures(
        self, rotation_matrix, rotation_centre=np.zeros((3, 1), dtype=np.float32)
    ):
        """
        rotates the components of the structure around a common point, default is the origin

        Parameters
        ----------
        rotation_matrix : open3d rotation matrix
            o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector)
        rotation_centre : 1*3 numpy float array
            centre of rotation for the structures

        Returns
        --------
        None
        """

        for item in range(len(self.solids)):
            self.solids[item] = GF.open3drotate(
                self.solids[item], rotation_matrix, rotation_centre
            )

    def translate_structures(self, vector):
        """
        translates the structures in the class by the given cartesian vector (x,y,z)

        Parameters
        -----------
        vector : 1*3 numpy array of floats
            The desired translation vector for the structures

        Returns
        --------
        None
        """
        for item in range(len(self.solids)):
            self.solids[item].translate(vector,relative=True)

    def export_vertices(self, structure_index=None):
        """
        Exports the vertices for either all or the indexed point clouds, transformed to the global coordinate frame.

        Parameters
        ----------
        structure_index : list
            list of structures of interest for vertices export

        Returns
        -------
        point_cloud :
        """
        point_cloud = o3d.geometry.PointCloud()
        if structure_index == None:
            # if no structure index is provided, then generate an index including all items in the class
            structure_index = []
            for item in range(len(self.solids)):
                structure_index.append(item)

        for item in structure_index:
            if self.solids[item] == None:
                # do nothing
                temp_cloud = o3d.geometry.PointCloud()
            else:
                temp_cloud = o3d.geometry.PointCloud()
                temp_cloud.points = self.solids[item].vertices
                if self.solids[item].has_vertex_normals():
                    temp_cloud.normals = self.solids[item].vertex_normals
                else:
                    self.solids[item].compute_vertex_normals()
                    temp_cloud.normals = self.solids[item].vertex_normals

                point_cloud = point_cloud + temp_cloud

        point_cloud.transform(self.pose)
        return point_cloud

    def import_cad(self,posixpath,scale=1.0,center=np.zeros((3,1))):
        """
        Import trianglemesh from cad format, with scalling factor is requried. Open3d supports .ply, .stl, .obj, .off, and .gltf/.glb files
        Parameters
        ----------
        posixpath
            file address to cad file to import
        scale
            scalling factor to account for units which are not SI.

        Returns
        -------

        """
        new_solid=o3d.io.read_triangle_mesh(posixpath.as_posix())
        new_solid.compute_triangle_normals()
        new_solid.scale(scale,center=center)
        self.add_structure(new_solid)
    def triangles_base_raycaster(self):
        """
        generates the triangles for all the :class:`open3d.geometry.TriangleMesh` objects in the structure, and outputs them as a continuous array of
        triangle_t format triangles

        Parameters
        -----------
        None

        Returns
        --------
        triangles : N by 1 numpy array of triangle_t triangles
            a continuous array of all the triangles in the structure
        """
        triangles = np.empty((0), dtype=base_types.triangle_t)
        for item in range(len(self.solids)):
            temp_object = copy.deepcopy(self.solids[item])
            temp_object.transform(self.pose)
            triangles = np.append(triangles, RF.convertTriangles(temp_object))

        return triangles


class antenna_structures(object3d):
    """
    Dedicated class to store information on a specific antenna, including aperture points
    as :class:`open3d.geometry.PointCloud` data structures, and structure shapes
    as :class:`open3d.geometry.TriangleMesh` data structures. Everything in the class will be considered an integrated
    unit, rotating and moving together. This inherits functions from the structures and points classes.

    This class will be developed to include material parameters to enable more complex modelling.

    Units should be SI, metres

    """

    def __init__(self, structures, points):
        super().__init__()
        self.structures = structures
        self.points = points
        self.antenna_xyz=o3d.geometry.TriangleMesh.create_coordinate_frame(
    size=1, origin=self.pose[:3,3]
)


    def rotate_antenna(
        self, rotation_matrix, rotation_centre=np.zeros((3, 1), dtype=np.float32)
    ):
        """
        rotates the components of the structure around a common point, default is the origin within the local coordinate system

        Parameters
        ----------
        rotation_matrix : open3d rotation matrix
            o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector)
        rotation_centre : 1*3 numpy float array
            centre of rotation for the structures

        Returns
        --------
        None
        """
        self.antenna_xyz=GF.open3drotate(self.antenna_xyz,rotation_matrix, rotation_centre
                )
        #translate to the rotation centre, then rotate around origin
        temp_origin=self.pose[:3,3].ravel()-rotation_centre.ravel()
        temp_origin=np.matmul(rotation_matrix,temp_origin)
        self.pose[:3,3]=temp_origin.ravel()+rotation_centre.ravel()
        self.pose[:3,:3]=np.matmul(rotation_matrix,self.pose[:3,:3])
        for item in range(len(self.structures.solids)):
            if (self.structures.solids[item] is not None):
                self.structures.solids[item] = GF.open3drotate(
                    self.structures.solids[item], rotation_matrix, rotation_centre
                )
        for item in range(len(self.points.points)):
            if (self.points.points[item] is not None):
                self.points.points[item] = GF.open3drotate(
                    self.points.points[item], rotation_matrix, rotation_centre
                )

    def translate_antenna(self, vector):
        """
        translates the structures and points in the class by the given cartesian vector (x,y,z) within the local coordinate system

        Parameters
        -----------
        vector : 1*3 numpy array of floats
            The desired translation vector for the antenna

        Returns
        --------
        None
        """

        self.pose[:3,3]=self.pose[:3,3].ravel()+vector.ravel()
        #for item in range(len(self.structures.solids)):
        #    self.structures.solids[item].translate(vector)
        #for item in range(len(self.points.points)):
        #    self.points.points[item].translate(vector)

    def export_all_points(self,point_index=None):
        if point_index==None:
            point_cloud = self.points.export_points()
            point_cloud.transform(self.pose) ##+ self.structures.export_vertices()
        else:
            point_cloud=self.points.export_points(point_index=point_index)
            point_cloud.transform(self.pose)

        return point_cloud

    def excitation_function(self,desired_e_vector,transmit_amplitude=1,point_index=None,phase_shift="none",wavelength=1.0,steering_vector = np.zeros((1, 3))):
        #generate the local excitation function and then convert into the global coordinate frame.
        if point_index==None:
            aperture_points=self.export_all_points()
        else:
            aperture_points = self.export_all_points(point_index=point_index)
        aperture_weights = EM.calculate_conformalVectors(
            desired_e_vector, np.asarray(aperture_points.normals), self.pose[:3,:3]
        )
        if phase_shift=="wavefront":
            source_points = np.asarray(aperture_points.points)
            phase_weights= BM.WavefrontWeights(source_points, steering_vector, wavelength)
            aperture_weights=aperture_weights*phase_weights.reshape(-1,1)

        return aperture_weights*transmit_amplitude

    def receive_transform(self,aperture_polarisation,excitation_function,beamforming_weights):
        #combine local aperture polarisation, received excitation function, and beamforming weights to calculate the received signal
        #convert global polarisation functions to local ones
        transform_R = R.from_matrix(self.pose[:3, :3])
        transform_global_to_local = transform_R.inv()
        local_signal = np.matmul(excitation_function, transform_global_to_local.as_matrix())
        local_received_signal = np.zeros((excitation_function.shape[1], 3))
        received_signal = np.zeros((excitation_function.shape[1], 3))
        local_received_signal[:, 0] = np.dot(beamforming_weights, local_signal[:, :, 0])
        local_received_signal[:, 1] = np.dot(beamforming_weights, local_signal[:, :, 1])
        local_received_signal[:, 2] = np.dot(beamforming_weights, local_signal[:, :, 2])
        received_signal = np.dot(local_received_signal, aperture_polarisation.ravel())

        return received_signal

    def export_all_structures(self):

        objects = copy.deepcopy(self.structures.solids)
        for item in range(len(objects)):
            if objects[item] is None:
                print("Structure does not exist")
            else:
                objects[item]=objects[item].transform(self.pose)

        return objects
    def farfield_distance(self, freq):
        # calculate farfield distance for the antenna, based upon the Fraunhofer distance
        total_points = self.export_all_points()
        # calculate bounding box
        bounding_box = total_points.get_oriented_bounding_box()
        max_points = bounding_box.get_max_bound()
        min_points = bounding_box.get_min_bound()
        center = bounding_box.get_center()
        max_dist = np.sqrt(
            (max_points[0] - center[0]) ** 2
            + (max_points[1] - center[1]) ** 2
            + (max_points[2] - center[2]) ** 2
        )
        min_dist = np.sqrt(
            (center[0] - min_points[0]) ** 2
            + (center[1] - min_points[1]) ** 2
            + (center[2] - min_points[2]) ** 2
        )
        a = np.mean([max_dist, min_dist])
        wavelength = 3e8 / freq
        farfield_distance = (2 * (2 * a) ** 2) / wavelength

        return farfield_distance

    def estimate_maximum_directivity(self, freq):
        # estimate the maximum possible directivity based upon the smallest enclosing sphere and Hannan's relation between projected area and directivity
        total_points = self.export_all_points()
        # calculate bounding box
        bounding_box = total_points.get_oriented_bounding_box()
        max_points = bounding_box.get_max_bound()
        min_points = bounding_box.get_min_bound()
        center = bounding_box.get_center()
        max_dist = np.sqrt(
            (max_points[0] - center[0]) ** 2
            + (max_points[1] - center[1]) ** 2
            + (max_points[2] - center[2]) ** 2
        )
        min_dist = np.sqrt(
            (center[0] - min_points[0]) ** 2
            + (center[1] - min_points[1]) ** 2
            + (center[2] - min_points[2]) ** 2
        )
        a = np.mean([max_dist, min_dist])
        projected_area = np.pi * (a ** 2)
        wavelength = 3e8 / freq
        directivity = (4 * np.pi * projected_area) / (wavelength ** 2)
        return directivity

    def visualise_antenna(self,extras=[],origin=True):
        """
        This function uses open3d to display the antenna structure in the local coordinate frame
        """
        total_points = self.export_all_points()
        # calculate bounding box
        if np.asarray(total_points.points).shape[0]<4:
            max_dist=2.0
            min_dist=0.0
        else:
            bounding_box = total_points.get_oriented_bounding_box()
            max_points = bounding_box.get_max_bound()
            min_points = bounding_box.get_min_bound()
            center = bounding_box.get_center()
            max_dist = np.sqrt(
                (max_points[0] - center[0]) ** 2
                + (max_points[1] - center[1]) ** 2
                + (max_points[2] - center[2]) ** 2
            )
            min_dist = np.sqrt(
                (center[0] - min_points[0]) ** 2
                + (center[1] - min_points[1]) ** 2
                + (center[2] - min_points[2]) ** 2
            )
        a = np.mean([max_dist, min_dist])
        #scale antenna xyz to the structures
        self.antenna_xyz=o3d.geometry.TriangleMesh.create_coordinate_frame(size=0.5*a, origin=[0, 0, 0])
        if origin:
            o3d.visualization.draw(
                self.points.points + self.structures.solids + [self.antenna_xyz] + extras
            )
        else:
            o3d.visualization.draw(
                self.points.points + self.structures.solids + extras
            )

    # def generate_farfield(self,excitation_vector,wavelength,elements=False,azimuth_resolution=37,elevation_resolution=37):
    #
    #
    #       if elements==False:
    #           resultant_pattern=antenna_pattern(azimuth_resolution=azimuth_resolution,elevation_resolution=elevation_resolution)
    #           resultant_pattern.pattern[:,:,0], resultant_pattern.pattern[:,:,1] =farfield_generator(self.points.export_points(),
    #                                                self.structures,
    #                                                excitation_vector,
    #                                                az_range=np.linspace(-180, 180, resultant_pattern.azimuth_resolution),
    #                                                el_range=np.linspace(-90, 90, resultant_pattern.elevation_resolution),
    #                                                wavelength=wavelength,
    #                                                farfield_distance=20,
    #                                                project_vectors=True,
    #                                                antenna_axes=self.antenna_axes,elements=elements)
    #       else:
    #           #generate antenna array pattern
    #           array_points=self.points.export_points()
    #           num_elements=np.asarray(np.asarray(array_points.points).shape[0])
    #           resultant_pattern = array_pattern(elements=num_elements,azimuth_resolution=azimuth_resolution,elevation_resolution=elevation_resolution)
    #           resultant_pattern.pattern[:,:, :, 0], resultant_pattern.pattern[:,:, :, 1] = farfield_generator(
    #               self.points.export_points(),
    #               self.structures,
    #               excitation_vector,
    #               az_range=np.linspace(-180, 180, resultant_pattern.azimuth_resolution),
    #               el_range=np.linspace(-90, 90, resultant_pattern.elevation_resolution),
    #               wavelength=wavelength,
    #               farfield_distance=20,
    #               project_vectors=True,
    #               antenna_axes=self.antenna_axes, elements=elements)
    #
    #       return resultant_pattern
    #
    # def frequency_domain_channel(self,sink_coords,excitation_function,wavelength=1.0,scatter_points=None,scattering=1,elements=True,):
    #
    #      Ex,Ey,Ez=frequency_domain_channel(self.points.export_points(),
    #                                    sink_coords,self.structures,excitation_function,scatter_points,wavelength,scattering,elements,project_vectors=True,antenna_axes=self.antenna_axes)
    #      return Ex,Ey,Ez

class antenna_pattern(object3d):
    """
    Antenna Pattern class which allows for patterns to be handled consistently
    across LyceanEM and other modules. The definitions assume that the pattern axes
    are consistent with the global axes set. If a different orientation is required,
    such as a premeasured antenna in a new orientation then the pattern rotate_function
    must be used.

    Antenna Pattern Frequency is in Hz
    Rotation Offset is Specified in terms of rotations around the x, y, and z axes as roll,pitch/elevation, and azimuth
    in radians.
    """

    def __init__(
        self,
        azimuth_resolution=37,
        elevation_resolution=37,
        pattern_frequency=1e9,
        arbitary_pattern=False,
        arbitary_pattern_type="isotropic",
        arbitary_pattern_format="Etheta/Ephi",
        file_location=None
    ):
        super().__init__()

        if file_location!=None:
            self.pattern_frequency = pattern_frequency
            self.arbitary_pattern_format = arbitary_pattern_format
            self.import_pattern(file_location)
            self.field_radius = 1.0
        else:
            self.azimuth_resolution = azimuth_resolution
            self.elevation_resolution = elevation_resolution
            self.pattern_frequency = pattern_frequency
            self.arbitary_pattern_type = arbitary_pattern_type
            self.arbitary_pattern_format = arbitary_pattern_format
            self.field_radius = 1.0
            az_mesh, elev_mesh = np.meshgrid(
                np.linspace(-180, 180, self.azimuth_resolution),
                np.linspace(-90, 90, self.elevation_resolution),
            )
            self.az_mesh = az_mesh
            self.elev_mesh = elev_mesh
            if self.arbitary_pattern_format == "Etheta/Ephi":

                self.pattern = np.zeros(
                    (self.elevation_resolution, self.azimuth_resolution, 2),
                    dtype=np.complex64,
                )

            elif self.arbitary_pattern_format == "ExEyEz":

                self.pattern = np.zeros(
                    (self.elevation_resolution, self.azimuth_resolution, 3),
                    dtype=np.complex64,
                )

            if arbitary_pattern == True:
                self.initilise_pattern()

    def initilise_pattern(self):
        """
        pattern initialisation function, providing an isotopic pattern
        or quasi-isotropic pattern

        Returns
        -------
        Populated antenna pattern
        """

        if self.arbitary_pattern_type == "isotropic":
            self.pattern[:, :, 0] = 1.0
        elif self.arbitary_pattern_type == "xhalfspace":
            az_angles = self.az_mesh[0, :]
            az_index = np.where(np.abs(az_angles) < 90)
            self.pattern[:, az_index, 0] = 1.0
        elif self.arbitary_pattern_type == "yhalfspace":
            az_angles = self.az_mesh[0, :]
            az_index = np.where(az_angles > 90)
            self.pattern[:, az_index, 0] = 1.0
        elif self.arbitary_pattern_type == "zhalfspace":
            elev_angles = self.elev_mesh[:, 0]
            elev_index = np.where(elev_angles > 0)
            self.pattern[elev_index, :, 0] = 1.0



    def import_pattern(self, file_location):
        """
        takes the file location and imports the individual pattern file, replacing exsisting values with those of the saved file.

        It is import to note that for CST ASCII export format, that you select a plot range of -180 to 180 for phi, and that by defauly CST exports from 0 to 180 in theta, which
        is the opposite direction to the default for LyceanEM, so this data structures are flipped for consistency.

        Parameters
        -----------
        file location : :class:`pathlib.Path`
            file location

        Returns
        ---------
        None
        """
        if file_location.suffix == ".txt":
            # file is a CST ffs format
            datafile = np.loadtxt(file_location, skiprows=2)
            theta = datafile[:, 0]
            phi = datafile[:, 1]
            freq_index1 = file_location.name.find("f=")
            freq_index2 = file_location.name.find(")")
            file_frequency = (
                float(file_location.name[freq_index1 + 2 : freq_index2]) * 1e9
            )
            phi_steps = np.linspace(np.min(phi), np.max(phi), np.unique(phi).size)
            theta_steps = np.linspace(
                np.min(theta), np.max(theta), np.unique(theta).size
            )
            phi_res = np.unique(phi).size
            theta_res = np.unique(theta).size
            etheta = (
                (datafile[:, 3] * np.exp(1j * np.deg2rad(datafile[:, 4])))
                .reshape(phi_res, theta_res)
                .transpose()
            )
            ephi = (
                (datafile[:, 5] * np.exp(1j * np.deg2rad(datafile[:, 6])))
                .reshape(phi_res, theta_res)
                .transpose()
            )
            self.azimuth_resolution = phi_res
            self.elevation_resolution = theta_res
            self.elev_mesh = np.flipud(
                GF.thetatoelevation(theta).reshape(phi_res, theta_res).transpose()
            )
            self.az_mesh = np.flipud(phi.reshape(phi_res, theta_res).transpose())
            self.pattern = np.zeros(
                (self.elevation_resolution, self.azimuth_resolution, 2),
                dtype=np.complex64,
            )
            self.pattern[:, :, 0] = np.flipud(etheta)
            self.pattern[:, :, 1] = np.flipud(ephi)
            self.pattern_frequency = file_frequency
        elif file_location.suffix == ".dat":
            # file is .dat format from anechoic chamber measurements
            Ea, Eb, freq, norm, theta_values, phi_values = EM.importDat(file_location)
            az_mesh, elev_mesh = np.meshgrid(
                phi_values, GF.thetatoelevation(theta_values)
            )
            self.azimuth_resolution = np.unique(phi_values).size
            self.elevation_resolution = np.unique(theta_values).size
            self.az_mesh = az_mesh
            self.elev_mesh = elev_mesh
            self.pattern = np.zeros(
                (self.elevation_resolution, self.azimuth_resolution, 2),
                dtype=np.complex64,
            )
            self.pattern_frequency = freq * 1e6
            self.pattern[:, :, 0] = Ea.transpose() + norm
            self.pattern[:, :, 1] = Eb.transpose() + norm

    def export_pattern(self, file_location):
        """
        takes the file location and exports the pattern as a .dat file
        unfinished, must be in Etheta/Ephi format,

        Parameters
        -----------
        file_location : :class:`pathlib.Path`
            the path for the output file, including name

        Returns
        --------
        None
        """
        if self.arbitary_pattern_format == "ExEyEz":
            self.transmute_pattern()

        theta_flat = GF.elevationtotheta(self.elev_mesh).transpose().reshape(-1, 1)
        phi_mesh = self.az_mesh.transpose().reshape(-1, 1)
        planes = self.azimuth_resolution
        copolardb = 20 * np.log10(
            np.abs(self.pattern[:, :, 0].transpose().reshape(-1, 1))
        )
        copolarangle = np.degrees(
            np.angle(self.pattern[:, :, 0].transpose().reshape(-1, 1))
        )
        crosspolardb = 20 * np.log10(
            np.abs(self.pattern[:, :, 1].transpose().reshape(-1, 1))
        )
        crosspolarangle = np.degrees(
            np.angle(self.pattern[:, :, 1].transpose().reshape(-1, 1))
        )
        norm = np.nanmax(np.array([np.nanmax(copolardb), np.nanmax(crosspolardb)]))
        copolardb -= norm
        crosspolardb -= norm
        infoarray = np.array(
            [
                planes,
                np.min(self.az_mesh),
                np.max(self.az_mesh),
                norm,
                self.pattern_frequency / 1e6,
            ]
        ).reshape(1, -1)

        dataarray = np.concatenate(
            (
                theta_flat.reshape(-1, 1),
                copolardb.reshape(-1, 1),
                copolarangle.reshape(-1, 1),
                crosspolardb.reshape(-1, 1),
                crosspolarangle.reshape(-1, 1),
            ),
            axis=1,
        )
        outputarray = np.concatenate((infoarray, dataarray), axis=0)
        np.savetxt(file_location, outputarray, delimiter=",", fmt="%.2e")

    def export_global_weights(self):
        """
        Calculates the global frame coordinates and xyz weightings of the antenna pattern using the pattern pose to translate and rotate into the correction position and orientation within the model environment. If this function is used as a source for models, then the antenna pattern pose must be updated for the desired position and orientation before export.

        Returns
        -------
        points :

        weights :

        """
        xyz = self.cartesian_points()
        points = o3d.geometry.PointCloud()
        points.points = o3d.utility.Vector3dVector(xyz)
        points.normals = o3d.utility.Vector3dVector(xyz)
        points.transform(self.pose)

        if self.arbitary_pattern_format == "Etheta/Ephi":
            self.transmute_pattern(desired_format="ExEyEz")

        #import to export the weights in cartesian format, then orientate correctly within global reference frame.
        rotation_matrix = self.pose[:3, :3]

        #weightsx= self.pattern[:,:,0].ravel()
        #weightsy = self.pattern[:, :, 1].ravel()
        #weightsz = self.pattern[:, :, 2].ravel()
        weights=np.array([self.pattern[:,:,0].reshape(-1),self.pattern[:,:,1].reshape(-1),self.pattern[:,:,2].reshape(-1)])
        weights=np.matmul(weights,rotation_matrix)
        return points, weights

    def display_pattern(
        self, plottype="Polar", desired_pattern="both", pattern_min=-40, plot_max=0
    ):
        """
        Displays the Antenna Pattern using :func:`lyceanem.electromagnetics.beamforming.PatternPlot`

        Parameters
        ----------
        plottype : str
            the plot type, either [Polar], [Cartesian-Surf], or [Contour]. The default is [Polar]
        desired_pattern : str
            the desired pattern, default is [both], but is Pattern format is 'Etheta/Ephi' then options are [Etheta] or [Ephi], and if Pattern format is 'ExEyEz', then options are [Ex], [Ey], or [Ez].
        pattern_min : float
            the desired scale minimum in dB, the default is [-40]

        Returns
        -------
        None
        """

        if self.arbitary_pattern_format == "Etheta/Ephi":
            if desired_pattern == "both":
                BM.PatternPlot(
                    self.pattern[:, :, 0],
                    self.az_mesh,
                    self.elev_mesh,
                    pattern_min=pattern_min,
                    plottype=plottype,
                    title_text="Etheta",
                    plot_max=plot_max,
                )
                BM.PatternPlot(
                    self.pattern[:, :, 1],
                    self.az_mesh,
                    self.elev_mesh,
                    pattern_min=pattern_min,
                    plottype=plottype,
                    title_text="Ephi",
                    plot_max=plot_max,
                )
            elif desired_pattern == "Etheta":
                BM.PatternPlot(
                    self.pattern[:, :, 0],
                    self.az_mesh,
                    self.elev_mesh,
                    pattern_min=pattern_min,
                    plottype=plottype,
                    title_text="Etheta",
                    plot_max=plot_max,
                )
            elif desired_pattern == "Ephi":
                BM.PatternPlot(
                    self.pattern[:, :, 1],
                    self.az_mesh,
                    self.elev_mesh,
                    pattern_min=pattern_min,
                    plottype=plottype,
                    title_text="Ephi",
                    plot_max=plot_max,
                )
            elif desired_pattern =="Power":
                BM.PatternPlot(
                    self.pattern[:, :, 0]**2+self.pattern[:, :, 1]**2,
                    self.az_mesh,
                    self.elev_mesh,
                    pattern_min=pattern_min,
                    logtype='power',
                    plottype=plottype,
                    title_text="Power Pattern",
                    plot_max=plot_max,
                )
        elif self.arbitary_pattern_format == "ExEyEz":
            if desired_pattern == "both":
                BM.PatternPlot(
                    self.pattern[:, :, 0],
                    self.az_mesh,
                    self.elev_mesh,
                    pattern_min=pattern_min,
                    plottype=plottype,
                    title_text="Ex",
                    plot_max=plot_max,
                )
                BM.PatternPlot(
                    self.pattern[:, :, 1],
                    self.az_mesh,
                    self.elev_mesh,
                    pattern_min=pattern_min,
                    plottype=plottype,
                    title_text="Ey",
                    plot_max=plot_max,
                )
                BM.PatternPlot(
                    self.pattern[:, :, 2],
                    self.az_mesh,
                    self.elev_mesh,
                    pattern_min=pattern_min,
                    plottype=plottype,
                    title_text="Ez",
                    plot_max=plot_max,
                )
            elif desired_pattern == "Ex":
                BM.PatternPlot(
                    self.pattern[:, :, 0],
                    self.az_mesh,
                    self.elev_mesh,
                    pattern_min=pattern_min,
                    plottype=plottype,
                    title_text="Ex",
                    plot_max=plot_max,
                )
            elif desired_pattern == "Ey":
                BM.PatternPlot(
                    self.pattern[:, :, 1],
                    self.az_mesh,
                    self.elev_mesh,
                    pattern_min=pattern_min,
                    plottype=plottype,
                    title_text="Ey",
                    plot_max=plot_max,
                )
            elif desired_pattern == "Ez":
                BM.PatternPlot(
                    self.pattern[:, :, 2],
                    self.az_mesh,
                    self.elev_mesh,
                    pattern_min=pattern_min,
                    plottype=plottype,
                    title_text="Ez",
                    plot_max=plot_max,
                )

    def transmute_pattern(self,desired_format="Etheta/Ephi"):
        """
        convert the pattern from Etheta/Ephi format to Ex, Ey,Ez format, or back again
        """
        if self.arbitary_pattern_format == "Etheta/Ephi":
            if desired_format=="ExEyEz":
                oldformat = self.pattern.reshape(-1, self.pattern.shape[2])
                old_shape = oldformat.shape
                self.arbitary_pattern_format = "ExEyEz"
                self.pattern = np.zeros(
                    (self.elevation_resolution, self.azimuth_resolution, 3),
                    dtype=np.complex64,
                )
                theta = GF.elevationtotheta(self.elev_mesh)
                # convrsion part, move to transmute pattern
                conversion_matrix1 = np.asarray(
                    [
                        np.cos(np.deg2rad(theta.ravel()))
                        * np.cos(np.deg2rad(self.az_mesh.ravel())),
                        np.cos(np.deg2rad(theta.ravel()))
                        * np.sin(np.deg2rad(self.az_mesh.ravel())),
                        -np.sin(np.deg2rad(theta.ravel())),
                    ]
                ).transpose()
                conversion_matrix2 = np.asarray(
                    [
                        -np.sin(np.deg2rad(self.az_mesh.ravel())),
                        np.cos(np.deg2rad(self.az_mesh.ravel())),
                        np.zeros(self.az_mesh.size),
                    ]
                ).transpose()
                decomposed_fields = (
                    oldformat[:, 0].reshape(-1, 1) * conversion_matrix1
                    + oldformat[:, 1].reshape(-1, 1) * conversion_matrix2
                )
                self.pattern[:, :, 0] = decomposed_fields[:, 0].reshape(
                    self.elevation_resolution, self.azimuth_resolution
                )
                self.pattern[:, :, 1] = decomposed_fields[:, 1].reshape(
                    self.elevation_resolution, self.azimuth_resolution
                )
                self.pattern[:, :, 2] = decomposed_fields[:, 2].reshape(
                    self.elevation_resolution, self.azimuth_resolution
                )
            elif desired_format == "Circular":
                #recalculate pattern using Right Hand Circular and Left Hand Circular Polarisation
                self.arbitary_pattern_format = desired_format

        else:
            if desired_format == "Etheta/Ephi":
                oldformat = self.pattern.reshape(-1, self.pattern.shape[2])
                old_shape = oldformat.shape
                self.arbitary_pattern_format = "Etheta/Ephi"
                self.pattern = np.zeros(
                    (self.elevation_resolution, self.azimuth_resolution, 2),
                    dtype=np.complex64,
                )
                theta = GF.elevationtotheta(self.elev_mesh)
                costhetacosphi = (
                    np.cos(np.deg2rad(self.az_mesh.ravel()))
                    * np.cos(np.deg2rad(theta.ravel()))
                ).astype(np.complex64)
                sinphicostheta = (
                    np.sin(np.deg2rad(self.az_mesh.ravel()))
                    * np.cos(np.deg2rad(theta.ravel()))
                ).astype(np.complex64)
                sintheta = (np.sin(np.deg2rad(theta.ravel()))).astype(np.complex64)
                sinphi = (np.sin(np.deg2rad(self.az_mesh.ravel()))).astype(np.complex64)
                cosphi = (np.cos(np.deg2rad(self.az_mesh.ravel()))).astype(np.complex64)
                new_etheta = (
                    oldformat[:, 0] * costhetacosphi
                    + oldformat[:, 1] * sinphicostheta
                    - oldformat[:, 2] * sintheta
                )
                new_ephi = -oldformat[:, 0] * sinphi + oldformat[:, 1] * cosphi
                self.pattern[:, :, 0] = new_etheta.reshape(
                    self.elevation_resolution, self.azimuth_resolution
                )
                self.pattern[:, :, 1] = new_ephi.reshape(
                    self.elevation_resolution, self.azimuth_resolution
                )

    def cartesian_points(self):
        """
        exports the cartesian points for all pattern points.
        """
        x, y, z = MF.sph2cart(
            np.deg2rad(self.az_mesh.ravel()),
            np.deg2rad(self.elev_mesh.ravel()),
            np.ones((self.pattern[:, :, 0].size)) * self.field_radius,
        )
        field_points = (
            np.array([x, y, z]).transpose().astype(np.float32)
        )

        return field_points

    def rotate_pattern(self, rotation_matrix=None):
        """
        Rotate the self pattern from the assumed global axes into the new direction

        Parameters
        ----------
        new_axes : 3x3 numpy float array
            the new vectors for the antenna x,y,z axes

        Returns
        -------
        Updates self.pattern with the new pattern reflecting the antenna
        orientation within the global models

        """
        # generate pattern_coordinates for rotation
        theta = GF.elevationtotheta(self.elev_mesh)
        x, y, z = RF.sph2cart(
            np.deg2rad(self.az_mesh.ravel()),
            np.deg2rad(self.elev_mesh.ravel()),
            np.ones((self.pattern[:, :, 0].size)),
        )
        field_points = np.array([x, y, z]).transpose().astype(np.float32)

        # convert to ExEyEz for rotation
        if self.arbitary_pattern_format == "Etheta/Ephi":
            self.transmute_pattern(desired_format="ExEyEz")
            desired_format = "Etheta/Ephi"
        else:
            desired_format = "ExEyEz"

        rotated_points = np.dot(field_points, rotation_matrix)
        decomposed_fields = self.pattern.reshape(-1, 3)
        # rotation part
        xyzfields = np.dot(decomposed_fields, rotation_matrix)
        # resample
        resampled_xyzfields = self.resample_pattern(
            rotated_points, xyzfields, field_points
        )
        self.pattern[:, :, 0] = resampled_xyzfields[:, 0].reshape(
            self.elevation_resolution, self.azimuth_resolution
        )
        self.pattern[:, :, 1] = resampled_xyzfields[:, 1].reshape(
            self.elevation_resolution, self.azimuth_resolution
        )
        self.pattern[:, :, 2] = resampled_xyzfields[:, 2].reshape(
            self.elevation_resolution, self.azimuth_resolution
        )

        if desired_format == "Etheta/Ephi":
            # convert back
            self.transmute_pattern(desired_format=desired_format)

    def resample_pattern_angular(
        self, new_azimuth_resolution, new_elevation_resolution
    ):
        """
        resample pattern based upon provided azimuth and elevation resolution
        """
        new_az_mesh, new_elev_mesh = np.meshgrid(
            np.linspace(-180, 180, new_azimuth_resolution),
            np.linspace(-90, 90, new_elevation_resolution),
        )
        x, y, z = RF.sph2cart(
            np.deg2rad(new_az_mesh.ravel()),
            np.deg2rad(new_elev_mesh.ravel()),
            np.ones((self.pattern[:, :, 0].size)),
        )
        new_field_points = np.array([x, y, z]).transpose().astype(np.float32)
        old_field_points = self.cartesian_points()
        old_pattern = self.pattern.reshape(-1, 2)
        new_pattern = self.resample_pattern(
            old_field_points, old_pattern, new_field_points
        )

    def resample_pattern(self, old_points, old_pattern, new_points):
        """

        Parameters
        ----------
        old_points : float xyz
            xyz coordinates that the pattern has been sampled at
        old_pattern : 2 or 3 by n complex array of the antenna pattern at the old_poitns
            DESCRIPTION.
        new_points : desired_grid points in xyz float array
            DESCRIPTION.

        Returns
        -------
        new points, new_pattern

        """
        pol_format = old_pattern.shape[-1]  # will be 2 or three
        new_pattern = np.zeros((new_points.shape[0], pol_format), dtype=np.complex64)
        smoothing_factor = 4
        for component in range(pol_format):
            mag_interpolate = sp.Rbf(
                old_points[:, 0],
                old_points[:, 1],
                old_points[:, 2],
                np.abs(old_pattern[:, component]),
                smooth=smoothing_factor,
            )
            phase_interpolate = sp.Rbf(
                old_points[:, 0],
                old_points[:, 1],
                old_points[:, 2],
                np.angle(old_pattern[:, component]),
                smooth=smoothing_factor,
            )
            new_mag = mag_interpolate(
                new_points[:, 0], new_points[:, 1], new_points[:, 2]
            )
            new_angles = phase_interpolate(
                new_points[:, 0], new_points[:, 1], new_points[:, 2]
            )
            new_pattern[:, component] = new_mag * np.exp(1j * new_angles)

        return new_pattern

    def directivity(self):
        """

        Returns
        -------
        Dtheta : numpy array
            directivity for Etheta farfield
        Dphi : numpy array
            directivity for Ephi farfield
        Dtotal : numpy array
            overall directivity pattern
        Dmax : numpy array
            the maximum directivity for each pattern

        """
        Dtheta, Dphi, Dtotal, Dmax = BM.directivity_transformv2(
            self.pattern[:, :, 0],
            self.pattern[:, :, 1],
            az_range=self.az_mesh[0, :],
            elev_range=self.elev_mesh[:, 0],
        )
        return Dtheta, Dphi, Dtotal, Dmax

class array_pattern:
    """
    Array Pattern class which allows for patterns to be handled consistently
    across LyceanEM and other modules. The definitions assume that the pattern axes
    are consistent with the global axes set. If a different orientation is required,
    such as a premeasured antenna in a new orientation then the pattern rotate_function
    must be used.

    Antenna Pattern Frequency is in Hz
    Rotation Offset is Specified in terms of rotations around the x, y, and z axes as roll,pitch/elevation, and azimuth
    in radians.
    """

    def __init__(
        self,
        azimuth_resolution=37,
        elevation_resolution=37,
        pattern_frequency=1e9,
        arbitary_pattern=False,
        arbitary_pattern_type="isotropic",
        arbitary_pattern_format="Etheta/Ephi",
        position_mapping=np.zeros((3), dtype=np.float32),
        rotation_offset=np.zeros((3), dtype=np.float32),
        elements=2,
    ):

        self.azimuth_resolution = azimuth_resolution
        self.elevation_resolution = elevation_resolution
        self.pattern_frequency = pattern_frequency
        self.arbitary_pattern_type = arbitary_pattern_type
        self.arbitary_pattern_format = arbitary_pattern_format
        self.position_mapping = position_mapping
        self.rotation_offset = rotation_offset
        self.field_radius = 1.0
        self.elements=elements
        self.beamforming_weights=np.ones((self.elements),dtype=np.complex64)
        az_mesh, elev_mesh = np.meshgrid(
            np.linspace(-180, 180, self.azimuth_resolution),
            np.linspace(-90, 90, self.elevation_resolution),
        )
        self.az_mesh = az_mesh
        self.elev_mesh = elev_mesh
        if self.arbitary_pattern_format == "Etheta/Ephi":
            self.pattern = np.zeros(
                ( elements,self.elevation_resolution, self.azimuth_resolution, 2),
                dtype=np.complex64,
            )
        elif self.arbitary_pattern_format == "ExEyEz":

            self.pattern = np.zeros(
                ( elements,self.elevation_resolution, self.azimuth_resolution, 3),
                dtype=np.complex64,
            )
        if arbitary_pattern == True:
            self.initilise_pattern()

    def _rotation_matrix(self):
        """_rotation_matrix getter method

        Calculates and returns the (3D) axis rotation matrix.

        Returns
        -------
        : :class:`numpy.ndarray` of shape (3, 3)
            The model (3D) rotation matrix.
        """
        x_rot = R.from_euler("X", self.rotation_offset[0], degrees=True).as_matrix()
        y_rot = R.from_euler("Y", self.rotation_offset[1], degrees=True).as_matrix()
        z_rot = R.from_euler("Z", self.rotation_offset[2], degrees=True).as_matrix()

        # x_rot=np.array([[1,0,0],
        #                [0,np.cos(np.deg2rad(self.rotation_offset[0])),-np.sin(np.deg2rad(self.rotation_offset[0]))],
        #                [0,np.sin(np.deg2rad(self.rotation_offset[0])),np.cos(np.deg2rad(self.rotation_offset[0]))]])

        total_rotation = np.dot(np.dot(z_rot, y_rot), x_rot)
        return total_rotation

    def initilise_pattern(self):
        """
        pattern initialisation function, providing an isotopic pattern
        or quasi-isotropic pattern

        Returns
        -------
        Populated antenna pattern
        """
        if self.arbitary_pattern_type == "isotropic":
            self.pattern[:, :,:, 0] = 1.0
        elif self.arbitary_pattern_type == "xhalfspace":
            az_angles = self.az_mesh[0, :]
            az_index = np.where(np.abs(az_angles) < 90)
            self.pattern[:, az_index,:, 0] = 1.0
        elif self.arbitary_pattern_type == "yhalfspace":
            az_angles = self.az_mesh[0, :]
            az_index = np.where(az_angles > 90)
            self.pattern[:, az_index, :,0] = 1.0
        elif self.arbitary_pattern_type == "zhalfspace":
            elev_angles = self.elev_mesh[:, 0]
            elev_index = np.where(elev_angles > 0)
            self.pattern[elev_index, :, :,0] = 1.0

    def display_pattern(
        self, plottype="Polar", desired_pattern="both", pattern_min=-40
    ):
        """
        Displays the Antenna Array Pattern using :func:`lyceanem.electromagnetics.beamforming.PatternPlot` and the stored weights

        Parameters
        ----------
        plottype : str
            the plot type, either [Polar], [Cartesian-Surf], or [Contour]. The default is [Polar]
        desired_pattern : str
            the desired pattern, default is [both], but is Pattern format is 'Etheta/Ephi' then options are [Etheta] or [Ephi], and if Pattern format is 'ExEyEz', then options are [Ex], [Ey], or [Ez].
        pattern_min : float
            the desired scale minimum in dB, the default is [-40]

        Returns
        -------
        None
        """

        if self.arbitary_pattern_format == "Etheta/Ephi":
            if desired_pattern == "both":
                BM.PatternPlot(
                    self.beamforming_weights*self.pattern[:,:, :, 0],
                    self.az_mesh,
                    self.elev_mesh,
                    pattern_min=pattern_min,
                    plottype=plottype,
                    title_text="Etheta",
                )
                BM.PatternPlot(
                    self.beamforming_weights*self.pattern[:,:, :, 1],
                    self.az_mesh,
                    self.elev_mesh,
                    pattern_min=pattern_min,
                    plottype=plottype,
                    title_text="Ephi",
                )
            elif desired_pattern == "Etheta":
                BM.PatternPlot(
                    self.beamforming_weights*self.pattern[:,:, :, 0],
                    self.az_mesh,
                    self.elev_mesh,
                    pattern_min=pattern_min,
                    plottype=plottype,
                    title_text="Etheta",
                )
            elif desired_pattern == "Ephi":
                BM.PatternPlot(
                    self.beamforming_weights*self.pattern[:,:, :, 1],
                    self.az_mesh,
                    self.elev_mesh,
                    pattern_min=pattern_min,
                    plottype=plottype,
                    title_text="Ephi",
                )
            elif desired_pattern =="Power":
                BM.PatternPlot(
                    (self.beamforming_weights*self.pattern[:,:, :, 0])**2+(self.beamforming_weights*self.pattern[:,:, :, 1])**2,
                    self.az_mesh,
                    self.elev_mesh,
                    pattern_min=pattern_min,
                    plottype=plottype,
                    title_text="Power Pattern",
                )
        elif self.arbitary_pattern_format == "ExEyEz":
            if desired_pattern == "both":
                BM.PatternPlot(
                    self.beamforming_weights*self.pattern[:,:, :, 0],
                    self.az_mesh,
                    self.elev_mesh,
                    pattern_min=pattern_min,
                    plottype=plottype,
                    title_text="Ex",
                )
                BM.PatternPlot(
                    self.beamforming_weights*self.pattern[:,:, :, 1],
                    self.az_mesh,
                    self.elev_mesh,
                    pattern_min=pattern_min,
                    plottype=plottype,
                    title_text="Ey",
                )
                BM.PatternPlot(
                    self.beamforming_weights*self.pattern[:,:, :, 2],
                    self.az_mesh,
                    self.elev_mesh,
                    pattern_min=pattern_min,
                    plottype=plottype,
                    title_text="Ez",
                )
            elif desired_pattern == "Ex":
                BM.PatternPlot(
                    self.beamforming_weights*self.pattern[:,:, :, 0],
                    self.az_mesh,
                    self.elev_mesh,
                    pattern_min=pattern_min,
                    plottype=plottype,
                    title_text="Ex",
                )
            elif desired_pattern == "Ey":
                BM.PatternPlot(
                    self.beamforming_weights*self.pattern[:,:, :, 1],
                    self.az_mesh,
                    self.elev_mesh,
                    pattern_min=pattern_min,
                    plottype=plottype,
                    title_text="Ey",
                )
            elif desired_pattern == "Ez":
                BM.PatternPlot(
                    self.beamforming_weights*self.pattern[:,:, :, 2],
                    self.az_mesh,
                    self.elev_mesh,
                    pattern_min=pattern_min,
                    plottype=plottype,
                    title_text="Ez",
                )


    def cartesian_points(self):
        """
        exports the cartesian points for all pattern points.
        """
        x, y, z = RF.sph2cart(
            np.deg2rad(self.az_mesh.ravel()),
            np.deg2rad(self.elev_mesh.ravel()),
            np.ones((self.pattern[:, :, 0].size)) * self.field_radius,
        )
        field_points = (
            np.array([x, y, z]).transpose().astype(np.float32)
        )

        return field_points

    def directivity(self):
        """

        Returns
        -------
        Dtheta : numpy array
            directivity for Etheta farfield
        Dphi : numpy array
            directivity for Ephi farfield
        Dtotal : numpy array
            overall directivity pattern
        Dmax : numpy array
            the maximum directivity for each pattern

        """
        Dtheta, Dphi, Dtotal, Dmax = BM.directivity_transformv2(
            self.beamforming_weights*self.pattern[:,:, :, 0],
            self.beamforming_weights*self.pattern[:,:, :, 1],
            az_range=self.az_mesh[0, :],
            elev_range=self.elev_mesh[:, 0],
        )
        return Dtheta, Dphi, Dtotal, Dmax



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Satellite Channel Modelling
======================================================

This example uses the frequency domain :func:`lyceanem.models.frequency_domain.calculate_scattering` function to
predict the scattering parameters for the frequency and environment included in the model. This allows the effects of antenna polarisation,


"""

# %%
# Define Map and Ground Station Location and list of Satellites using skyfield API to load starlink satellites.
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

import folium

interest_center = [51.458729, -2.602153]  # location of Bristol Physics Observatory

m = folium.Map(
    location=[
        interest_center[0],
        interest_center[1],
    ],
    zoom_start=6,
    tiles="Stamen Terrain",
)

import numpy as np
import copy

from skyfield.api import load, EarthSatellite, wgs84

BristolPhysicsObservatory = wgs84.latlon(interest_center[0], interest_center[1], 0)

receiver = BristolPhysicsObservatory

# Build the time range `t` over which to plot, plus other values.

ts = load.timescale()

time_points = 2 * 60  # track over 10 minutes
# change over time
t = ts.utc(2022, 12, 15, 13, 4, range(0, time_points))
timebase = t.utc_datetime()

# %%
# GeoJSON
# ~~~~~~~~
# The Folium plugin `TimestampedGeoJson`_ will be used to plot our tracks using timestamped
# GeoJSONs. As a result, we want to convert our data into `GeoJSON format`_. Firstly, we create
# our feature collection which we will append our features to.
#
# .. _GeoJSON format: https://geojson.org/
# .. _TimestampedGeoJson:  https://python-visualization.github.io/folium/plugins.html

geo_features = list()
geo_json = {
    "type": "FeatureCollection",
    "features": geo_features,
}
# %%
# Plotting Tracks of Satellites visible to Physics Observatory
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Now we append our tracks to our feature collection list. We define `colour_iter` which will
# allow us to cycle through different colours when mapping our tracks.

from collections import defaultdict
from itertools import cycle

colour_iter = iter(
    cycle(
        [
            "red",
            "blue",
            "green",
            "purple",
            "orange",
            "darkred",
            "#0909FF",
            "#F70D1A",
            "#FF6700",
            "lightgreen",
            "#0AFFFF",
            "#12AD2B",
            "#E2F516",
            "#FFFF00",
            "#F52887",
        ]
    )
)
colour = defaultdict(lambda: next(colour_iter))

trail_Size = 7  # trail_Size is the number of timestamps we want track to trail for
from io import BytesIO
import base64


# %%
# Define Starlinks Tracks
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Convinience function to limit the tracks to a set number of the nearest N satellites.

def starlink_tracks(pointofinterest, timebase, num_satellites=5):
    starlink_url = "https://celestrak.com/NORAD/elements/supplemental/starlink.txt"
    starlinks = load.tle_file(starlink_url)
    tracks = []
    for sat in starlinks:
        pos = sat.at(timebase)
        tracks.append(pos)
        latcheck = np.any(np.isnan(wgs84.subpoint_of(pos).latitude.degrees))
        loncheck = np.any(np.isnan(wgs84.subpoint_of(pos).longitude.degrees))
        if latcheck | loncheck:
            tracks.pop(-1)
            print("caught one")

    distance_trim = np.full((len(tracks), len(timebase)), np.inf)
    for track in range(len(tracks)):
        lat = wgs84.subpoint_of(tracks[track]).latitude
        lon = wgs84.subpoint_of(tracks[track]).longitude
        relative_measure = (starlinks[track] - pointofinterest).at(timebase)
        alt, az, distance1 = relative_measure.altaz()
        altitude = alt.degrees
        azimuth = az.degrees
        relative_range = distance1.km
        plot_points = []
        plot_times = []
        az_trim = []
        alt_trim = []
        range_trim = []
        for point in range(len(lat.degrees)):
            if altitude[point] >= 0.0:
                distance_trim[track, point] = relative_range[point]

    def get_indices_of_k_smallest(arr, k):
        idx = np.argpartition(arr.ravel(), k)
        return tuple(np.array(np.unravel_index(idx, arr.shape))[:, range(min(k, 0), max(k, 0))])


    indexholder = np.zeros((num_satellites, len(timebase)), dtype=int)
    for point in range(len(timebase)):
        indexholder[:, point] = get_indices_of_k_smallest(distance_trim[:, point], num_satellites)[0]

    return starlinks, indexholder

starlinks, indexholder = starlink_tracks(BristolPhysicsObservatory, t, num_satellites=20)

tracks = []
for sat in starlinks:
    pos = sat.at(t)
    tracks.append(pos)
    latcheck = np.any(np.isnan(wgs84.subpoint_of(pos).latitude.degrees))
    loncheck = np.any(np.isnan(wgs84.subpoint_of(pos).longitude.degrees))
    if latcheck | loncheck:
        tracks.pop(-1)
        print("caught one")

for track in range(len(tracks)):
    lat = wgs84.subpoint_of(tracks[track]).latitude
    lon = wgs84.subpoint_of(tracks[track]).longitude
    relative_measure = (starlinks[track] - BristolPhysicsObservatory).at(t)
    alt, az, distance = relative_measure.altaz()
    altitude = alt.degrees
    azimuth = az.degrees
    relative_range = distance.km
    plot_points = []
    plot_times = []
    az_trim = []
    alt_trim = []
    range_trim = []
    for point in range(len(lat.degrees)):
        if altitude[point] >= 0.0 and np.any(np.isin(track, indexholder[:, point])):
            plot_points.append((lon.degrees[point], lat.degrees[point]))
            plot_times.append(timebase[point].strftime("%Y-%m-%d %H:%M:%S"))
            az_trim.append(azimuth[point])
            alt_trim.append(altitude[point])
            range_trim.append(relative_range[point])

    for time_index, time in enumerate(plot_times):
        geo_features.append(
            {
                "type": "Feature",
                "properties": {
                    "name": starlinks[track].name,
                    "style": {"color": colour[track], "weight": 6},
                    "times": [time] * len(plot_points[: time_index + 1][-trail_Size:]),
                },
                "geometry": {
                    "type": "LineString",
                    "coordinates": plot_points[: time_index + 1][-trail_Size:],
                },
            }
        )
        geo_features.append(
            {
                "type": "Feature",
                "properties": {
                    "icon": "marker",
                    "iconstyle": {
                        "iconUrl": f"http://icons.iconarchive.com/icons/google/noto-emoji-travel-places/1024/42597-satellite-icon.png",
                        "iconSize": [24, 24],
                        "fillOpacity": 1,
                        "popupAnchor": [1, -17],
                    },
                    "popup": "Satellite: " + starlinks[track].name + "<dd>"
                                                                     "Latitude: "
                             + "%s" % float("%.8g" % plot_points[time_index][0])
                             + "<dd>"
                               "Longitude: "
                             + "%s" % float("%.8g" % plot_points[time_index][1])
                             + "<dd>"  # rounding 8 sigfigs
                               "Azimuth: "
                             + "%s" % float("%.8g" % az_trim[time_index])
                             + "°"
                             + "<dd>"
                               "Altitude: "
                             + "%s" % float("%.8g" % alt_trim[time_index])
                             + "°"
                             + "<dd>"
                               "Slant Range: "
                             + "%s" % float("%.8g" % range_trim[time_index])
                             + "km",
                    "name": starlinks[track].name,
                    "style": {"color": "black", "weight": 2},
                    "times": [time],
                },
                "geometry": {
                    "type": "MultiPoint",
                    "coordinates": [plot_points[time_index]],
                },
            }
        )

# plot receiver
r_track = []
for time in range(len(t)):
    pos = BristolPhysicsObservatory.at(t)
    r_track.append(pos)

for track in range(len(r_track)):
    lat = wgs84.subpoint_of(r_track[track]).latitude
    lon = wgs84.subpoint_of(r_track[track]).longitude
    relative_measure = (receiver - BristolPhysicsObservatory).at(t)
    alt, az, distance = relative_measure.altaz()
    altitude = alt.degrees
    azimuth = az.degrees
    relative_range = distance.km
    plot_points = []
    plot_times = []
    az_trim = []
    alt_trim = []
    range_trim = []
    for point in range(len(lat.degrees)):
        # if altitude[point] >= 0.0:
        plot_points.append((lon.degrees[point], lat.degrees[point]))
        plot_times.append(timebase[point].strftime("%Y-%m-%d %H:%M:%S"))
        az_trim.append(azimuth[point])
        alt_trim.append(altitude[point])
        range_trim.append(relative_range[point])

    for time_index, time in enumerate(plot_times):
        geo_features.append(
            {
                "type": "Feature",
                "properties": {
                    "name": "Ground Station at Bristol Physics Observatory",
                    "style": {"color": colour[track], "weight": 6},
                    "times": [time] * len(plot_points[: time_index + 1][-trail_Size:]),
                },
                "geometry": {
                    "type": "LineString",
                    "coordinates": plot_points[: time_index + 1][-trail_Size:],
                },
            }
        )
        geo_features.append(
            {
                "type": "Feature",
                "properties": {
                    "icon": "marker",
                    "iconstyle": {
                        "iconUrl": f"https://www.bristol.ac.uk/media-library/sites/physics/new-images/501_10013623.jpg",
                        "iconSize": [64, 32],
                        "fillOpacity": 1,
                        "popupAnchor": [1, -17],
                    },
                    "popup": "Ground Station at Bristol Physics Observatory" + "<dd>"
                                                                               "Latitude: "
                             + "%s" % float("%.8g" % plot_points[time_index][0])
                             + "<dd>"
                               "Longitude: "
                             + "%s" % float("%.8g" % plot_points[time_index][1])
                             + "<dd>"  # rounding 8 sigfigs
                               "Azimuth: "
                             + "%s" % float("%.8g" % az_trim[time_index])
                             + "°"
                             + "<dd>"
                               "Altitude: "
                             + "%s" % float("%.8g" % alt_trim[time_index])
                             + "°"
                             + "<dd>"
                               "Slant Range: "
                             + "%s" % float("%.8g" % range_trim[time_index])
                             + "km",
                    "name": "Ground Station at Bristol Physics Observatory",
                    "style": {"color": "black", "weight": 2},
                    "times": [time],
                },
                "geometry": {
                    "type": "MultiPoint",
                    "coordinates": [plot_points[time_index]],
                },
            }
        )

from folium.plugins import TimestampedGeoJson, Fullscreen

Fullscreen().add_to(m)

import numpy as np
import matplotlib.pyplot as plt
from geopandas import GeoDataFrame
from shapely.geometry import Polygon, MultiPolygon


def collec_to_gdf(collec_poly):
    """Transform a `matplotlib.contour.QuadContourSet` to a GeoDataFrame"""
    polygons, colors = [], []
    for i, polygon in enumerate(collec_poly.collections):
        mpoly = []
        for path in polygon.get_paths():
            try:
                path.should_simplify = False
                poly = path.to_polygons()
                # Each polygon should contain an exterior ring + maybe hole(s):
                exterior, holes = [], []
                if len(poly) > 0 and len(poly[0]) > 3:
                    # The first of the list is the exterior ring :
                    exterior = poly[0]
                    # Other(s) are hole(s):
                    if len(poly) > 1:
                        holes = [h for h in poly[1:] if len(h) > 3]
                mpoly.append(Polygon(exterior, holes))
            except:
                print('Warning: Geometry error when making polygon #{}'
                      .format(i))
        if len(mpoly) > 1:
            mpoly = MultiPolygon(mpoly)
            polygons.append(mpoly)
            colors.append(polygon.get_facecolor().tolist()[0])
        elif len(mpoly) == 1:
            polygons.append(mpoly[0])
            colors.append(polygon.get_facecolor().tolist()[0])
    return GeoDataFrame(
        geometry=polygons,
        data={'RGBA': colors},
        crs={'init': 'epsg:4326'})

# %%
# The Results
# ~~~~~~~~~~~~~~

TimestampedGeoJson(
    data=geo_json,
    transition_time=200,
    auto_play=True,
    add_last_point=False,
    period="PT1S",
    duration="PT0S",
).add_to(m)

# %%


m.show_in_browser()








#!/usr/bin/env python
# coding: utf-8
"""
Modelling a Physical Channel in the Frequency Domain
======================================================

This example uses the frequency domain :func:`lyceanem.models.frequency_domain.calculate_scattering` function to
predict the scattering parameters for the frequency and environment included in the model.
This model allows for a very wide range of antennas and antenna arrays to be considered, but for simplicity only horn
antennas will be included in this example. The simplest case would be a single source point and single receive point,
rather than an aperture antenna such as a horn.



"""

import numpy as np
import open3d as o3d
import copy

# %%
# Frequency and Mesh Resolution
# ------------------------------
#
freq = np.asarray(15.0e9)
wavelength = 3e8 / freq
mesh_resolution = 0.5 * wavelength

# %%
# Setup transmitters and receivers
# -----------------------------------
#
import lyceanem.geometry.targets as TL
import lyceanem.geometry.geometryfunctions as GF

transmit_horn_structure, transmitting_antenna_surface_coords = TL.meshedHorn(
    58e-3, 58e-3, 128e-3, 2e-3, 0.21, mesh_resolution
)
receive_horn_structure, receiving_antenna_surface_coords = TL.meshedHorn(
    58e-3, 58e-3, 128e-3, 2e-3, 0.21, mesh_resolution
)

# %%
# Position Transmitter
# ----------------------
# rotate the transmitting antenna to the desired orientation, and then translate to final position.
# :func:`lyceanem.geometryfunctions.open3drotate` allows both the center of rotation to be defined, and ensures the
# right syntax is used for Open3d, as it was changed from 0.9.0 to 0.10.0 and onwards.
#
rotation_vector1 = np.radians(np.asarray([90.0, 0.0, 0.0]))
rotation_vector2 = np.radians(np.asarray([0.0, 0.0, -90.0]))
transmit_horn_structure = GF.open3drotate(
    transmit_horn_structure,
    o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector1),
)
transmit_horn_structure = GF.open3drotate(
    transmit_horn_structure,
    o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector2),
)
transmit_horn_structure.translate(np.asarray([2.695, 0, 0]), relative=True)
transmitting_antenna_surface_coords = GF.open3drotate(
    transmitting_antenna_surface_coords,
    o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector1),
)
transmitting_antenna_surface_coords = GF.open3drotate(
    transmitting_antenna_surface_coords,
    o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector2),
)
transmitting_antenna_surface_coords.translate(np.asarray([2.695, 0, 0]), relative=True)
# %%
# Position Receiver
# ------------------
# rotate the receiving horn to desired orientation and translate to final position.
receive_horn_structure = GF.open3drotate(
    receive_horn_structure,
    o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector1),
)
receive_horn_structure.translate(np.asarray([0, 1.427, 0]), relative=True)
receiving_antenna_surface_coords = GF.open3drotate(
    receiving_antenna_surface_coords,
    o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector1),
)
receiving_antenna_surface_coords.translate(np.asarray([0, 1.427, 0]), relative=True)

# %%
# Create Scattering Plate
# --------------------------
# Create a Scattering plate a source of multipath reflections

reflectorplate, scatter_points = TL.meshedReflector(
    0.3, 0.3, 6e-3, wavelength * 0.5, sides="front"
)
position_vector = np.asarray([29e-3, 0.0, 0])
rotation_vector1 = np.radians(np.asarray([0.0, 90.0, 0.0]))
scatter_points = GF.open3drotate(
    scatter_points,
    o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector1),
)
reflectorplate = GF.open3drotate(
    reflectorplate,
    o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector1),
)
reflectorplate.translate(position_vector, relative=True)
scatter_points.translate(position_vector, relative=True)

# %%
# Specify Reflection Angle
# --------------------------
# Rotate the scattering plate to the optimum angle for reflection from the transmitting to receiving horn

plate_orientation_angle = 45.0

rotation_vector = np.radians(np.asarray([0.0, 0.0, plate_orientation_angle]))
scatter_points = GF.open3drotate(
    scatter_points,
    o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector),
)
reflectorplate = GF.open3drotate(
    reflectorplate,
    o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector),
)

from lyceanem.base_classes import structures

blockers = structures([reflectorplate, receive_horn_structure, transmit_horn_structure])

# %%
# Visualise the Scene Geometry
# ------------------------------
# Use open3d function :func:`open3d.visualization.draw_geometries` to visualise the scene and ensure that all the
# relavent sources and scatter points are correct. Point normal vectors can be displayed by pressing 'n' while the
# window is open.
mesh_frame = o3d.geometry.TriangleMesh.create_coordinate_frame(
    size=0.5, origin=[0, 0, 0]
)
o3d.visualization.draw_geometries(
    [
        transmitting_antenna_surface_coords,
        receiving_antenna_surface_coords,
        scatter_points,
        reflectorplate,
        mesh_frame,
        receive_horn_structure,
        transmit_horn_structure,
    ]
)
# %%
# .. image:: ../_static/03_frequency_domain_channel_model_picture_01.png
#

# %%
# Specify desired Transmit Polarisation
# --------------------------------------
# The transmit polarisation has a significant effect on the channel characteristics. In this example the transmit
# horn will be vertically polarised, (e-vector aligned with the y direction)

desired_E_axis = np.zeros((1, 3), dtype=np.float32)
desired_E_axis[0, 1] = 1.0

# %%
# Frequency Domain Scattering
# ----------------------------
# Once the arrangement of interest has been setup, :func:`lyceanem.models.frequency_domain.calculate_scattering` can
# be called, using raycasting to calculate the scattering parameters based upon the inputs. The scattering parameter
# determines how many reflections will be considered. A value of 0 would mean only line of sight contributions will be
# calculated, with 1 including single reflections, and 2 including double reflections as well.

import lyceanem.models.frequency_domain as FD

Ex, Ey, Ez = FD.calculate_scattering(
    aperture_coords=transmitting_antenna_surface_coords,
    sink_coords=receiving_antenna_surface_coords,
    antenna_solid=blockers,
    desired_E_axis=desired_E_axis,
    scatter_points=scatter_points,
    wavelength=wavelength,
    scattering=1,
)

# %%
# Examine Scattering
# ---------------------
# The resultant scattering is decomposed into the Ex,Ey,Ez components at the receiving antenna, by itself this is not
# that interesting, so for this example we will rotate the reflector back, and then create a loop to step the reflector
# through different angles from 0 to 90 degrees in 1 degree steps.


angle_values = np.linspace(0, 90, 91)
angle_increment = np.diff(angle_values)[0]
responsex = np.zeros((len(angle_values)), dtype="complex")
responsey = np.zeros((len(angle_values)), dtype="complex")
responsez = np.zeros((len(angle_values)), dtype="complex")

plate_orientation_angle = -45.0

rotation_vector = np.radians(
    np.asarray([0.0, 0.0, plate_orientation_angle + angle_increment])
)
scatter_points = GF.open3drotate(
    scatter_points,
    o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector),
)
reflectorplate = GF.open3drotate(
    reflectorplate,
    o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector),
)

from tqdm import tqdm

for angle_inc in tqdm(range(len(angle_values))):
    rotation_vector = np.radians(np.asarray([0.0, 0.0, angle_increment]))
    scatter_points = GF.open3drotate(
        scatter_points,
        o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector),
    )
    reflectorplate = GF.open3drotate(
        reflectorplate,
        o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector),
    )
    Ex, Ey, Ez = FD.calculate_scattering(
        aperture_coords=transmitting_antenna_surface_coords,
        sink_coords=receiving_antenna_surface_coords,
        antenna_solid=blockers,
        desired_E_axis=desired_E_axis,
        scatter_points=scatter_points,
        wavelength=wavelength,
        scattering=1,
    )
    responsex[angle_inc] = Ex
    responsey[angle_inc] = Ey
    responsez[angle_inc] = Ez

# %%
# Plot Normalised Response
# ----------------------------
# Using matplotlib, plot the results

import matplotlib.pyplot as plt

normalised_max = np.max(
    np.array(
        [
            np.max(20 * np.log10(np.abs(responsex))),
            np.max(20 * np.log10(np.abs(responsey))),
            np.max(20 * np.log10(np.abs(responsez))),
        ]
    )
)
ExdB = 20 * np.log10(np.abs(responsex)) - normalised_max
EydB = 20 * np.log10(np.abs(responsey)) - normalised_max
EzdB = 20 * np.log10(np.abs(responsez)) - normalised_max

fig, ax = plt.subplots()
ax.plot(angle_values - 45, ExdB, label="Ex")
ax.plot(angle_values - 45, EydB, label="Ey")
ax.plot(angle_values - 45, EzdB, label="Ez")
plt.xlabel("$\\theta_{N}$ (degrees)")
plt.ylabel("Normalised Level (dB)")
ax.set_ylim(-60.0, 0)
ax.set_xlim(np.min(angle_values) - 45, np.max(angle_values) - 45)
ax.set_xticks(np.linspace(np.min(angle_values) - 45, np.max(angle_values) - 45, 19))
ax.set_yticks(np.linspace(-60, 0.0, 21))
legend = ax.legend(loc="upper right", shadow=True)
plt.grid()
plt.show()



import pytest
import numpy as np
import open3d as o3d
from numpy.testing import assert_equal, assert_allclose
from scipy.spatial.transform import Rotation as R
from ..base_classes import antenna_structures,structures,points

def cube():
    # a cube centered at the origin, with side lengths of 1m (default)
    cube = o3d.geometry.TriangleMesh.create_box()
    cube.translate(np.array([-0.5,-0.5,-0.5]))
    return cube

def point():
    #a single point on the +x center of the cube with consistent normal vector
    source_point=o3d.geometry.PointCloud()
    source_point.points=o3d.utility.Vector3dVector(np.array([[0.5,0,0]]))
    source_point.normals=o3d.utility.Vector3dVector(np.array([[1.0, 0, 0]]))
    return source_point

def antenna():
    base = structures([cube()])
    source_point = points([point()])
    antenna = antenna_structures(base,source_point)
    return antenna

@pytest.fixture
def standard_antenna():
    return antenna()

def test_excitation_function_x_u(standard_antenna):
    #test that an unrotated antenna with u (horizontal-y) polarisation gives horizontal-y polarisation
    desired_E_vector=np.array([1,0,0],dtype=np.complex64)
    final_vector = np.array([[0, 1, 0]], dtype=np.complex64)
    assert_allclose(standard_antenna.excitation_function(desired_e_vector=desired_E_vector),final_vector,atol=1e-12)

def test_excitation_function_x_v(standard_antenna):
    #test that an unrotated antenna with u (horizontal-y) polarisation gives horizontal-y polarisation
    desired_E_vector=np.array([0,1,0],dtype=np.complex64)
    final_vector = np.array([[0, 0, 1]], dtype=np.complex64)
    assert_allclose(standard_antenna.excitation_function(desired_e_vector=desired_E_vector),final_vector,atol=1e-12)

def test_excitation_function_x_n(standard_antenna):
    #test that an unrotated antenna with u (horizontal-y) polarisation gives horizontal-y polarisation
    desired_E_vector=np.array([0,0,1],dtype=np.complex64)
    final_vector = np.array([[1, 0, 0]], dtype=np.complex64)
    assert_allclose(standard_antenna.excitation_function(desired_e_vector=desired_E_vector),final_vector,atol=1e-12)

def test_excitation_function_x_y_u(standard_antenna):
    #test that a rotated antenna (about z, rotating x to y direction with u (horizontal-y) polarisation gives horizontal-x polarisation
    desired_E_vector=np.array([1,0,0],dtype=np.complex64)
    final_vector=np.array([[-1,0,0]],dtype=np.complex64)
    standard_antenna.pose[:3,:3]= R.from_euler('z',90,degrees=True).as_matrix()
    assert_allclose(standard_antenna.excitation_function(desired_e_vector=desired_E_vector),final_vector,atol=1e-12)

def test_excitation_function_x_mz_u(standard_antenna):
    #test that a rotated antenna (about y, rotating x to z direction with u (horizontal-y) polarisation gives horizontal-x polarisation
    desired_E_vector=np.array([1,0,0],dtype=np.complex64)
    final_vector=np.array([[0,1,0]],dtype=np.complex64)
    standard_antenna.pose[:3,:3]= R.from_euler('y',90,degrees=True).as_matrix()
    assert_allclose(standard_antenna.excitation_function(desired_e_vector=desired_E_vector),final_vector,atol=1e-12)

def test_excitation_function_x_mz_v(standard_antenna):
    #test that a rotated antenna (about y, rotating x to z direction with u (horizontal-y) polarisation gives horizontal-x polarisation
    desired_E_vector=np.array([0,1,0],dtype=np.complex64)
    final_vector=np.array([[1,0,0]],dtype=np.complex64)
    standard_antenna.pose[:3,:3]= R.from_euler('y',90,degrees=True).as_matrix()
    assert_allclose(standard_antenna.excitation_function(desired_e_vector=desired_E_vector),final_vector,atol=1e-12)

def test_excitation_function_x_mz_cp(standard_antenna):
    #test that a rotated antenna (about y, rotating x to z direction with uv circular polarisation (xy) polarisation gives horizontal-x polarisation
    desired_E_vector=np.array([1*np.exp(-1j*0),1*np.exp(-1j*(np.pi/2)),0],dtype=np.complex64)
    final_vector=np.array([[1*np.exp(-1j*(np.pi/2)),1*np.exp(-1j*0),0]],dtype=np.complex64)
    standard_antenna.pose[:3,:3]= R.from_euler('y',90,degrees=True).as_matrix()
    assert_allclose(standard_antenna.excitation_function(desired_e_vector=desired_E_vector),final_vector,atol=1e-12)


import numpy as np

# provide idealised patterns to allow testing of the different models
import open3d as o3d

from lyceanem.base_classes import antenna_pattern
import lyceanem.geometry.geometryfunctions as GF



def electriccurrentsource(prime_vector, theta, phi):
    """
    create an idealised electric current source that can be used to test the outputs of the model

    Parameters
    ----------
    prime_vector : 1D numpy array of floats
        orientation of the electric current source in xyz
    theta : 2D numpy array of floats
        theta angles of desired pattern in degrees
    phi : 2D numpy array of floats
        phi angles of desired pattern in degrees

    Returns
    -------
    etheta : 2D numpy array of complex
        Etheta polarisation
    ephi : 2D numpy array of complex
        Ephi polarisation
    """

    etheta = np.zeros(theta.shape, dtype=np.complex64)
    ephi = np.zeros(theta.shape, dtype=np.complex64)
    etheta = (
        prime_vector[0] * np.cos(np.deg2rad(phi)) * np.cos(np.deg2rad(theta))
        + prime_vector[1] * np.sin(np.deg2rad(phi)) * np.cos(np.deg2rad(theta))
        - prime_vector[2] * np.sin(np.deg2rad(theta))
    )
    ephi = -prime_vector[0] * np.sin(np.deg2rad(phi)) + prime_vector[1] * np.cos(
        np.deg2rad(phi)
    )
    return etheta, ephi


def antenna_pattern_source(radius, import_antenna=False, antenna_file=None):
    """
    This function generates an antenna pattern and `opaque' sphere as the base, representing an inserted antenna with measured pattern.

    This function is not yet complete
    Parameters
    ----------
    radius : float
        radius of the sphere, setting the minimum enclosing volume of the antenna
    import_antenna : boolean,
        if [True] the provided antenna_file location will be used to import an antenna file to populate the variable
    antenna_file : PosixPath
        a file location for the antenna file to be used. The initial set will be based upon the .dat files used by the University of Bristol Anechoic Chamber

    Returns
    --------
    solid : :class:`open3d.geometry.TriangleMesh`
        the enclosing sphere for the antenna
    points : :class:`open3d.geometry.PointCloud`
        the sample points for the antenna pattern, to be used as source points for the frequency domain model
    pattern : 3 by N numpy array of complex
        array of the sample points of the antenna pattern, specified as Ex,Ey,Ez components

    """
    if import_antenna:
        # import antenna pattern
        pattern = antenna_pattern()
        pattern.import_pattern(antenna_file)
        pattern.field_radius = radius
    else:
        print("arbitary pattern")
        # generate an arbitary locally Z directed electric current source
        prime_vector = np.zeros((3), dtype=np.float32)
        prime_vector[2] = 1.0
        az_res = 37
        elev_res = 37
        az_mesh, elev_mesh = np.meshgrid(
            np.linspace(-180, 180, az_res), np.linspace(-90, 90, elev_res)
        )
        _, theta = np.meshgrid(
            np.linspace(-180.0, 180.0, az_res),
            GF.elevationtotheta(np.linspace(-90, 90, elev_res)),
        )
        etheta, ephi = electriccurrentsource(prime_vector, theta, az_mesh)
        pattern = antenna_pattern()
        pattern.pattern[:, :, 0] = etheta
        pattern.pattern[:, :, 1] = ephi
        pattern.field_radius = radius

    field_points = pattern.cartesian_points()
    points = o3d.geometry.PointCloud()
    points.points = o3d.utility.Vector3dVector(field_points)
    points.normals = o3d.utility.Vector3dVector(field_points)
    points.normalize_normals()
    solid = o3d.geometry.TriangleMesh.create_sphere(radius=radius, resolution=30)

    return solid, points, pattern



#!/usr/bin/env python
# coding: utf-8
"""
Modelling a Coherently Polarised Aperture
======================================================

This example uses the frequency domain :func:`lyceanem.models.frequency_domain.calculate_farfield` function to predict
the farfield pattern for a linearly polarised aperture. This could represent an antenna array without any beamforming
weights.


"""
import numpy as np
import open3d as o3d
import copy

# %%
# Setting Farfield Resolution and Wavelength
# -------------------------------------------
# LyceanEM uses Elevation and Azimuth to record spherical coordinates, ranging from -180 to 180 degrees in azimuth,
# and from -90 to 90 degrees in elevation. In order to launch the aperture projection function, the resolution in
# both azimuth and elevation is requried.
# In order to ensure a fast example, 37 points have been used here for both, giving a total of 1369 farfield points.
#
# The wavelength of interest is also an important variable for antenna array analysis, so we set it now for 10GHz,
# an X band aperture.

az_res = 181
elev_res = 181
wavelength = 3e8 / 10e9

# %%
# Geometries
# ------------------------
# In order to make things easy to start, an example geometry has been included within LyceanEM for a UAV, and the
# :class:`open3d.geometry.TriangleMesh` structures can be accessed by importing the data subpackage
import lyceanem.tests.reflectordata as data

body, array, source_coords = data.exampleUAV(10e9)

# %%
# Visualise the Resultant UAV and Array
# ---------------------------------------
# :func:`open3d.visualization.draw_geometries` can be used to visualise the open3d data
# structures :class:`open3d.geometry.PointCloud` and :class:`open3d.geometry.PointCloud`

mesh_frame = o3d.geometry.TriangleMesh.create_coordinate_frame(
    size=0.5, origin=[0, 0, 0]
)
o3d.visualization.draw_geometries([body, array, source_coords, mesh_frame])

# %%
# .. image:: ../_static/UAVArraywithPoints.png

# crop the inner surface of the array trianglemesh (not strictly required, as the UAV main body provides blocking to
# the hidden surfaces, but correctly an aperture will only have an outer face.
surface_array = copy.deepcopy(array)
surface_array.triangles = o3d.utility.Vector3iVector(
    np.asarray(array.triangles)[: len(array.triangles) // 2, :]
)
surface_array.triangle_normals = o3d.utility.Vector3dVector(
    np.asarray(array.triangle_normals)[: len(array.triangle_normals) // 2, :]
)

from lyceanem.base_classes import structures

blockers = structures([body, array])

from lyceanem.models.frequency_domain import calculate_farfield

from lyceanem.geometry.targets import source_cloud_from_shape

source_points, _ = source_cloud_from_shape(surface_array, wavelength * 0.5)

o3d.visualization.draw_geometries([body, array, source_points])

# %%
# .. image:: ../_static/sourcecloudfromshapeuav.png

# %%
# Drawbacks of :func:`lyceanem.geometry.geometryfunctions.sourcecloudfromshape`
# ------------------------------------------------------------------------------
# As can be seen by comparing the two source point sets, :func:`lyceanem.geometry.geometryfunctions.sourcecloudfromshape`
# has a significant drawback when used for complex sharply curved antenna arrays, as the poisson disk sampling method
# does not produce consistently spaced results.

desired_E_axis = np.zeros((1, 3), dtype=np.float32)
desired_E_axis[0, 2] = 1.0

Etheta, Ephi = calculate_farfield(
    source_coords,
    blockers,
    desired_E_axis,
    az_range=np.linspace(-180, 180, az_res),
    el_range=np.linspace(-90, 90, elev_res),
    wavelength=wavelength,
    farfield_distance=20,
    project_vectors=True,
)

# %%
# Storing and Manipulating Antenna Patterns
# ---------------------------------------------
# The resultant antenna pattern can be stored in :class:`lyceanem.base.antenna_pattern` as it has been modelled as one
# distributed aperture, the advantage of this class is the integrated display, conversion and export functions. It is
# very simple to define, and save the pattern, and then display with a call
# to :func:`lyceanem.base.antenna_pattern.display_pattern`. This produces 3D polar plots which can be manipulated to
# give a better view of the whole pattern, but if contour plots are required, then this can also be produced by passing
# plottype='Contour' to the function.

from lyceanem.base_classes import antenna_pattern

UAV_Static_Pattern = antenna_pattern(
    azimuth_resolution=az_res, elevation_resolution=elev_res
)
UAV_Static_Pattern.pattern[:, :, 0] = Etheta
UAV_Static_Pattern.pattern[:, :, 0] = Ephi

UAV_Static_Pattern.display_pattern()

# %%
# .. image:: ../_static/sphx_glr_02_coherently_polarised_array_001.png
# .. image:: ../_static/sphx_glr_02_coherently_polarised_array_002.png

UAV_Static_Pattern.display_pattern(plottype="Contour")

# %%
# .. image:: ../_static/sphx_glr_02_coherently_polarised_array_003.png
# .. image:: ../_static/sphx_glr_02_coherently_polarised_array_004.png
